/*

    File: main.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <scrnsave.h>
#include <gl/gl.h>
#include <math.h>
#include <stdio.h>

#include "main.h"
#include "points.h"
#include "texgen.h"

HDC hDC = NULL;
HGLRC hRC = NULL;
UINT timerID = 0;
int win_width = 100, win_height = 100;

bool fade_effect = true;
unsigned char fade_alpha = 63;	/* out of 255*/

bool using_window = false;

int requestExit = 0;

char *dialogClassName = "OpenGL";
char *windowName = "Screensaver By Chris Moore";
int diag_winWidth, diag_winHeight, diag_winX, diag_winY;

void DisplayScreen(void)
{
	GLbitfield gl_bits = GL_DEPTH_BUFFER_BIT;

	if (!fade_effect)
	{
		gl_bits |= GL_COLOR_BUFFER_BIT;
	}
	else
	{
		glPushAttrib(GL_ALL_ATTRIB_BITS);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
//		glDisable(GL_TEXTURE_2D);
glEnable(GL_TEXTURE_2D);
glDisable(GL_TEXTURE_GEN_S);
glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_DEPTH_TEST);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0,1,0,1,-1,0);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

//		glColor4ub(0,0,0,fade_alpha);
glColor4ub(1,1,1,fade_alpha);
		glBegin(GL_QUADS);
		glTexCoord2i(0,0);			glVertex2d(0,0);
		glTexCoord2i(0,1);			glVertex2d(0,1);
		glTexCoord2i(1,1);			glVertex2d(1,1);
		glTexCoord2i(1,0);			glVertex2d(1,0);
		glEnd();

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);

		glPopAttrib();
	}


	glClear(gl_bits);

	glLoadIdentity();

	DisplayPoints();

	glFlush();
	SwapBuffers(hDC);
}

#define ViewplaneDist		1
#define ViewplaneFarDist	1000

void ResizeOpenGL(int x, int y)
{
	float ratio = (float)y / (float)x;

	glViewport(0,0,x,y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-ViewplaneDist/ratio, ViewplaneDist/ratio,
			-ViewplaneDist, ViewplaneDist,
			ViewplaneDist, ViewplaneFarDist);


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void InitOpenGL(void)
{
	ResizeOpenGL(win_width, win_height);

	GenerateTexture();
}

int CALLBACK MyTimerProc(void)
{
	DisplayScreen();
	return 0;
}

void KillWindow(HWND hWnd)
{
	ShutdownPoints();

	if (timerID) KillTimer(hWnd, timerID);

	wglMakeCurrent(NULL, NULL);
	ReleaseDC(hWnd, hDC);
	wglDeleteContext(hRC);
}

bool InitWindow(HWND hWnd)
{	
	PIXELFORMATDESCRIPTOR pfd;
	int pf;

	hDC = GetDC(hWnd);

	/* there is no guarantee that the contents of the stack that become
	   the pfd are zeroed, therefore _make sure_ to clear these bits. */
	memset(&pfd, 0, sizeof(pfd));
	pfd.nSize		= sizeof(pfd);
	pfd.nVersion	= 1;
	pfd.dwFlags		= PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType	= PFD_TYPE_RGBA;
	pfd.cColorBits	= 32;

	pf = ChoosePixelFormat(hDC, &pfd);
	if (pf == 0) {
	MessageBox(NULL, "ChoosePixelFormat() failed:  "
		"Cannot find a suitable pixel format.", "Error", MB_OK); 
		return false;
	}
 
	if (SetPixelFormat(hDC, pf, &pfd) == FALSE) {
	MessageBox(NULL, "SetPixelFormat() failed:  "
		"Cannot set format specified.", "Error", MB_OK);
		return false;
	} 

	DescribePixelFormat(hDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

	hRC = wglCreateContext(hDC);
	wglMakeCurrent(hDC, hRC);

	if (using_window == false)
	{
		win_width = GetSystemMetrics(SM_CXSCREEN);
		win_height = GetSystemMetrics(SM_CYSCREEN);
	}
	else
	{
		win_width = diag_winWidth;
		win_height = diag_winHeight;
	}


	//randomize via the time:
	SYSTEMTIME systime;
	GetSystemTime((LPSYSTEMTIME)&systime);
	srand(systime.wMinute +
		systime.wSecond +
		systime.wMilliseconds);

	InitOpenGL();

	if (!InitPoints())
	{
		MessageBox(NULL, "Init Points Failed", "Error", MB_OK);
		return false;
	}

	if (using_window == false)
	{
		//if we're doing the screensaver then latch the update func to the timer
		if ((timerID = SetTimer(hWnd, 1, 1, (TIMERPROC)MyTimerProc)) == 0) return false;
		//else, just copy 'er over to the main loop
	}
	return true;
}

LRESULT WINAPI ScreenSaverProc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static PAINTSTRUCT ps;

	switch (message) {
  	case WM_ERASEBKGND:
		return 1;		//this should say "dont erase me"
	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;
	case WM_CREATE:
		using_window = false;
		if (InitWindow(hWnd) == false) PostQuitMessage(0);
		return 0;
	case WM_DESTROY:
		KillWindow(hWnd);
		PostQuitMessage(0);
		return 0;
	}

	// Deal with any unprocessed messages. . .
	return DefScreenSaverProc(hWnd, message, wParam, lParam);
}

BOOL WINAPI ScreenSaverConfigureDialog (HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	return true;
}

HMENU hMenu;
HMENU hFileMenu;
HMENU hOptionsMenu;
HMENU hAboutMenu;

LRESULT APIENTRY DialogWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static PAINTSTRUCT ps;

	switch (message) {
  	case WM_ERASEBKGND:
		return 1;		//this should say "dont erase me"
	case WM_PAINT:
		BeginPaint(hWnd, &ps);
		EndPaint(hWnd, &ps);
		return 0;
	case WM_CREATE:
		if (InitWindow(hWnd) == false) PostQuitMessage(0);
		return 0;
	case WM_SIZE:
		ResizeOpenGL(LOWORD(lParam),HIWORD(lParam));
		PostMessage(hWnd, WM_PAINT, 0, 0);
		return 0;

	case WM_COMMAND:
		{
			int wNotifyCode = HIWORD(wParam); // notification code 
			int wID = LOWORD(wParam);         // item, control, or accelerator identifier 
			HWND hwndCtl = (HWND) lParam;      // handle of control 

			if (wNotifyCode == 0)	//then a menu is being clicked and wID is the menu item identifier
			{
				switch (wID) {
				case 10:
					PostQuitMessage(0);
					break;
				}
			}
		}
		return 0;

	case WM_CLOSE:
		requestExit = 1;
		return 0;
    }

    return DefWindowProc(hWnd, message, wParam, lParam);
}


BOOL WINAPI RegisterDialogClasses (HANDLE hInst)
{
    WNDCLASS wndClass;
    HWND hWnd;
    MSG msg;

	using_window = true;

    // Define and register a window class
    wndClass.style = CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
    wndClass.lpfnWndProc = DialogWndProc;
    wndClass.cbClsExtra = 0;
    wndClass.cbWndExtra = 0;
    wndClass.hInstance = (HINSTANCE)hInst;
    wndClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wndClass.hCursor = LoadCursor(NULL, IDC_ARROW);
    wndClass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
    wndClass.lpszMenuName = NULL;
    wndClass.lpszClassName = dialogClassName;
    RegisterClass(&wndClass);

    // Figure out a default size for the window
    diag_winWidth = GetSystemMetrics(SM_CXSCREEN)/2;
    diag_winHeight = GetSystemMetrics(SM_CYSCREEN)/2;
	diag_winX = diag_winWidth/2;
	diag_winY = diag_winHeight/2;

    // Create a window of the previously defined class
    hWnd = CreateWindow(
        dialogClassName,        // Window class's name
        windowName,             // Title bar text
        WS_OVERLAPPEDWINDOW |   // The window's style
        WS_CLIPCHILDREN |
        WS_CLIPSIBLINGS,
        diag_winX, diag_winY,             // Position
        diag_winWidth, diag_winHeight,    // Size
        NULL,                   // Parent window's handle
        NULL,                   // Menu handle
        (HINSTANCE)hInst,                  // Instance handle
        NULL);                  // No additional data

    hDC = GetDC(hWnd);

	//try out a menu
	hMenu = CreateMenu();
	hFileMenu = CreateMenu();
	hOptionsMenu = CreateMenu();
	hAboutMenu = CreateMenu();

	InsertMenu(hMenu, 1, MF_BYPOSITION | MF_POPUP, (int)hFileMenu, "File");
	InsertMenu(hMenu, 2, MF_BYPOSITION | MF_POPUP, (int)hOptionsMenu, "Options");
	InsertMenu(hMenu, 3, MF_BYPOSITION | MF_POPUP, (int)hAboutMenu, "About");

	InsertMenu(hFileMenu, 1, MF_BYCOMMAND, 10, "Quit");

	//options:
	//display flags
	//use fade
	//fade level
	//use blend
	//use depth test
	//use texturing
	//camera distance
	InsertMenu(hOptionsMenu, 1, MF_BYCOMMAND, 11, "Dot");
	
	SetMenu(hWnd, hMenu);
	DrawMenuBar(hWnd);

    // Map the window to the screen
    ShowWindow(hWnd, SW_SHOW);

	while (!requestExit) {
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT) break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else {
			DisplayScreen();
		}
	}

    return msg.wParam;
}