/*

    File: points.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <windows.h>
#include <gl/gl.h>
#include <stdio.h>

#include <math.h>

#include "points.h"
#include "vector.h"
#include "main.h"
#include "texgen.h"	//something new - for texture_id


typedef struct t_obj            t_obj;
typedef struct t_vtx            t_vtx;
typedef struct t_particle       t_particle;

struct t_obj {
  vector        pos, vel, accel;
  vector		ang, rot;
} object;

struct t_vtx {
  float			color[4];
  vector        pos, newpos;
};
  //newpos is constantly changing, pos is constant

struct t_particle {
              //world, view
  vector        pos, vel, normal;
  float			color[4];
  long          life;
} *particle = NULL;

int particle_next;

long PARTICLE_MAX;
long PARTICLE_LIFE;
float PARTICLE_DIST;
float PARTICLE_SPEED;


vector camerapos(0,20,0);

int VERTEXMAX=0;
t_vtx *vertex = NULL;

int display_flags;

void ThreeDquad(t_particle *p1, t_particle *p2, t_particle *p3, t_particle *p4)
{
	glBegin(GL_QUADS);

	glNormal3fv(p1->normal.v);
	glTexCoord3fv(p1->pos.v);
	glColor4fv(p1->color);
	glVertex3fv(p1->pos.v);

	glNormal3fv(p2->normal.v);
	glTexCoord3fv(p2->pos.v);
	glColor4fv(p2->color);
	glVertex3fv(p2->pos.v);

	glNormal3fv(p3->normal.v);
	glTexCoord3fv(p3->pos.v);
	glColor4fv(p3->color);
	glVertex3fv(p3->pos.v);

	glNormal3fv(p4->normal.v);
	glTexCoord3fv(p4->pos.v);
	glColor4fv(p4->color);
	glVertex3fv(p4->pos.v);

	glEnd();
}

void ThreeDtri(t_particle *p1, t_particle *p2, t_particle *p3)
{
	glBegin(GL_TRIANGLES);

	glNormal3fv(p1->normal.v);
	glTexCoord3fv(p1->pos.v);
	glColor4fv(p1->color);
	glVertex3fv(p1->pos.v);

	glNormal3fv(p2->normal.v);
	glTexCoord3fv(p2->pos.v);
	glColor4fv(p2->color);
	glVertex3fv(p2->pos.v);

	glNormal3fv(p3->normal.v);
	glTexCoord3fv(p3->pos.v);
	glColor4fv(p3->color);
	glVertex3fv(p3->pos.v);

	glEnd();
}

void ThreeDtube(t_particle *src, t_particle *dest)
{
	int a;
	vector srcv[3], destv[3];

	srand((int)src);
	for (a=0;a<3;a++)	srcv[a] = vec_add(vector(RandomNum(), RandomNum(), RandomNum()),src->pos);

	srand((int)dest);
	for (a=0;a<3;a++)	destv[a] = vec_add(vector(RandomNum(), RandomNum(), RandomNum()),dest->pos);

	glBegin(GL_QUADS);
	for (a = 0; a < 3; a++)
	{
		int a_inc = (a+1)%3;

		glTexCoord3fv(srcv[a].v);
		glColor4fv(src->color);
		glVertex3fv(srcv[a].v);

		glTexCoord3fv(destv[a].v);
		glColor4fv(dest->color);
		glVertex3fv(destv[a].v);

		glTexCoord3fv(destv[a].v);
		glColor4fv(dest->color);
		glVertex3fv(destv[a_inc].v);

		glTexCoord3fv(srcv[a].v);
		glColor4fv(src->color);
		glVertex3fv(srcv[a_inc].v);
	}
	glEnd();
}

void ThreeDline (t_particle *src, t_particle *dest)
{
	glBegin(GL_LINES);

	glNormal3fv(src->normal.v);
	glTexCoord3fv(src->pos.v);
	glColor4fv(src->color);
	glVertex3fv(src->pos.v);

	glNormal3fv(dest->normal.v);
	glTexCoord3fv(dest->pos.v);
	glColor4fv(dest->color);
	glVertex3fv(dest->pos.v);

	glEnd();
}
	
void ThreeDflat(t_particle *src, t_particle *dest)
{
	vector q[4];
	q[0] = src->pos;
	q[1] = dest->pos;
	q[2] = vec_mul(dest->pos, 0.75f);
	q[3] = vec_mul(src->pos, 0.75f);

	static float vzero[4] = {0,0,0,0};

	float *color[4] = {src->color, dest->color, vzero, vzero};
	float tc[4][2] = {{0,0},{1,0},{1,1},{0,1}};

	float *n[4] = {src->normal.v, dest->normal.v, dest->normal.v, src->normal.v};

	glBegin(GL_QUADS);
	for (int a = 0; a < 4; a++)
	{
		glNormal3fv(n[a]);
		glTexCoord3fv(tc[a]);
		glColor4fv(color[a]);
		glVertex3fv(q[a].v);
	}
	glEnd();
}

void ThreeDdot(t_particle *p)
{
	glBegin(GL_POINTS);
	glNormal3fv(p->normal.v);
	glColor4fv(p->color);
	glVertex3fv(p->pos.v);
	glEnd();
}

#define DOT2_SIZE	1
void ThreeDdot2(t_particle *p)
{
	glColor4fv(p->color);
	glNormal3fv(p->normal.v);

	glBegin(GL_QUADS);

	glTexCoord2f(0,0);
	glVertex3f(	p->pos.x-DOT2_SIZE,
				p->pos.y-DOT2_SIZE,
				p->pos.z);

	glTexCoord2f(0,1);
	glVertex3f(	p->pos.x-DOT2_SIZE,
				p->pos.y+DOT2_SIZE,
				p->pos.z);

	glTexCoord2f(1,1);
	glVertex3f(	p->pos.x+DOT2_SIZE,
				p->pos.y+DOT2_SIZE,
				p->pos.z);

	glTexCoord2f(1,0);
	glVertex3f(	p->pos.x+DOT2_SIZE,
				p->pos.y-DOT2_SIZE,
				p->pos.z);

	glEnd();
}


void emit_particle(t_vtx *v, t_vtx *nextv)
{
	t_particle *p_new;

	p_new = &particle[particle_next];
	particle_next = (particle_next + 1) % PARTICLE_MAX;

	p_new->pos = vec_add(vec_mul(v->pos, PARTICLE_DIST), object.pos);
	p_new->vel = vec_mul(v->newpos, PARTICLE_SPEED);

	static float scalar_theta=0;
	scalar_theta += 0.0001f;
	p_new->vel.scale(.5f*(float)sin(scalar_theta + (float)(int)v)+.5f);

	vector d1, d2;
	d1 = vec_neg(v->newpos);
	d2 = vec_sub(nextv->newpos, v->newpos);
	p_new->normal.cross(&d1, &d2);
	p_new->normal.normalize();

	memcpy(p_new->color, v->color, sizeof(p_new->color));
	
	p_new->life = PARTICLE_LIFE;
}

void emit_particles(void)
{
	int a;
	for (a = 0; a < VERTEXMAX; a++)
		emit_particle(&vertex[a], &vertex[(a+1)%VERTEXMAX]);
}




void update_particles(void)
{
	t_particle *p, *last, *prev;//, *lastprev;

	for (int a = 0; a < PARTICLE_MAX; a++)
	{
		int last_a, prev_a;//, lastprev_a;
		last_a = (a - VERTEXMAX + PARTICLE_MAX) % PARTICLE_MAX;
		prev_a = ((a + VERTEXMAX - 1) % VERTEXMAX) + (a / VERTEXMAX) * VERTEXMAX;
//		lastprev_a = ((a + VERTEXMAX - 1) % VERTEXMAX) + ((a/VERTEXMAX + 1) % PARTICLE_LIFE) * VERTEXMAX
		p = &particle[a];
		last = &particle[last_a];
		prev = &particle[prev_a];

		for (int b = 0; b < 3; b++)
			p->pos.v[b] += p->vel.v[b];

		p->life -= 1;

		vector n = vec_normal(p->vel);

		p->color[3] = (float)p->life / (float)PARTICLE_LIFE;
		p->color[3] *= p->color[3];
		if (p->life > 0)
		{
//			ThreeDdot2(p);
//			ThreeDdot(p);
//			if (last->life > 0) ThreeDline (p, last);
			if (last->life > 0) ThreeDflat(p,last);
//			if (last->life > 0) ThreeDtube(p, last);
//			if (prev->life > 0) ThreeDline (p, prev);
//			if (last->life > 0 && prev->life > 0) ThreeDtri(p, last, prev);
//			if (prev->life > 0)	ThreeDtri(p, prev, particle);
//			if (last->life > 0 &&
//				prev->life > 0 &&
//				lastprev->life > 0)
//				ThreeDquad(p, last, lastprev, prev);
		}
	}
}


#define RANGE_MAX	40.f
void DisplayPoints(void)
{
	int a;

	{
		static double mtimer = 0;
		mtimer+=0.004;

		t_obj newobj;
		newobj = object;

		newobj.pos = vector(
			(float)sin(5.25*mtimer),
			(float)sin(4.69*mtimer),
			(float)sin(3.97*mtimer));
		newobj.pos.scale(RANGE_MAX);

		object = newobj;
	}

 //add rotation to angle
	for (a = 0; a < 3; a++)
	{
		object.ang.v[a] += object.rot.v[a];

		if (object.ang.v[a] > TWO_PI)
			object.ang.v[a] = (float)fmod((double)object.ang.v[a], TWO_PI);
		else
			object.ang.v[a] = (float)(TWO_PI - fmod(-(double)object.ang.v[a], TWO_PI));
	}

	for (a = 0; a < VERTEXMAX; a++)
		rotate(vertex[a].pos, &vertex[a].newpos, object.ang);   //rotate

/*
	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glScalef(1.f/256.f, 1.f/256.f, 1.f/256.f);
	glRotatef(object.ang.x*180.f/M_PI,1,0,0);
	glRotatef(object.ang.y*180.f/M_PI,0,1,0);
	glRotatef(object.ang.z*180.f/M_PI,0,0,1);
	glMatrixMode(GL_MODELVIEW);
*/

	glLoadIdentity();
	glTranslatef(-camerapos.x, -camerapos.y, -camerapos.z);

	update_particles();

	emit_particles();
}

// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
//SHUTDOWN


void ShutdownPoints(void)
{
	free(vertex);
	free(particle);
}


// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
// ***************************************************************************************
//INIT

void init_particle(void)
{
//all the data we will change:
	PARTICLE_LIFE = 500;
	PARTICLE_DIST = 0;
	PARTICLE_SPEED = .1f;
	object.pos = vector(0,0,0);
	object.vel = vector(.5,1,.25);
	object.accel = vector(0,0,0);
	object.ang = vector(0,0,0);
	object.rot = vector(
		DegtoRad(RandomNum() * 10.0f),
		DegtoRad(RandomNum() * 10.0f),
		DegtoRad(RandomNum() * 10.0f));


	object.ang = vector(0,0,0);
	PARTICLE_MAX = PARTICLE_LIFE * VERTEXMAX;

	particle = (t_particle *)malloc(PARTICLE_MAX * sizeof(t_particle));
	if (!particle) {
		MessageBox(NULL, "Failed to allocate particle buffer", "Error!", MB_OK); 
		PostQuitMessage(0);
	}
	particle_next = 0;
}

bool InitPoints(void)
{
	//generate the verticies
	VERTEXMAX = 10;
	vertex = (t_vtx *)malloc(VERTEXMAX * sizeof(t_vtx));
	if (!vertex) return false;
	memset(vertex, 0, VERTEXMAX * sizeof(t_vtx));

	for (int a = 0; a < VERTEXMAX; a++)
	{
		vertex[a].pos = vector(RandomNum(),RandomNum(),RandomNum());
		vertex[a].pos.normalize();
		/*
		vertex[a].color[0] = (float)(rand()&0xFF)/255.f;
		vertex[a].color[1] = (float)(rand()&0xFF)/255.f;
		vertex[a].color[2] = (float)(rand()&0xFF)/255.f;
		*/

		int i;
		vector colorvector;
		colorvector = vertex[a].pos;
		for (i = 0; i < 3; i++) colorvector.v[i] = vertex[a].pos.v[i] < 0 ? -vertex[a].pos.v[i] : vertex[a].pos.v[i];

		colorvector.y += 1.f;
		colorvector.z += 1.f;
		colorvector.normalize();

		for (i = 0; i < 3; i++) vertex[a].color[i] = colorvector.v[i];

		vertex[a].color[3] = 1.f;
	}


	init_particle();

	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

	glDisable(GL_TEXTURE_2D);
//	glEnable(GL_TEXTURE_2D);

	glEnable(GL_BLEND);
//	glEnable(GL_DEPTH_TEST);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);

//	glEnable(GL_TEXTURE_GEN_S);
//	glEnable(GL_TEXTURE_GEN_T);

  return true;
}
