/*

    File: vector.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <math.h>

#include "vector.h"

vector vec_rot_x (vector vec, float rot)
{
	return vector(
    vec.x,
    vec.y * (float)cos(rot) - vec.z * (float)sin(rot),
    vec.z * (float)cos(rot) + vec.y * (float)sin(rot));
}

vector vec_rot_y (vector vec, float rot)
{
  return vector(
    vec.x * (float)cos(rot) + vec.z * (float)sin(rot),
    vec.y,
    vec.z * (float)cos(rot) - vec.x * (float)sin(rot));
}

vector vec_rot_z (vector vec, float rot)
{
  return vector(
    vec.x * (float)cos(rot) - vec.y * (float)sin(rot),
    vec.y * (float)cos(rot) + vec.x * (float)sin(rot),
    vec.z);
}

void rotate (vector vec, vector *outvec, vector rot)
{
  vector v_sin, v_cos;
  vector temp;

  v_sin.x = (float)sin(rot.x);
  v_sin.y = (float)sin(rot.y);
  v_sin.z = (float)sin(rot.z);
  v_cos.x = (float)cos(rot.x);
  v_cos.y = (float)cos(rot.y);
  v_cos.z = (float)cos(rot.z);

  temp.y = vec.y * v_cos.x - vec.z * v_sin.x;
  temp.z = vec.z * v_cos.x + vec.y * v_sin.x;
  outvec->y = temp.y;
  outvec->z = temp.z;

  temp.z = outvec->z * v_cos.y - vec.x * v_sin.y;
  temp.x = vec.x * v_cos.y + outvec->z * v_sin.y;
  outvec->z = temp.z;
  outvec->x = temp.x;

  temp.x = outvec->x * v_cos.z - outvec->y * v_sin.z;
  temp.y = outvec->y * v_cos.z + outvec->x * v_sin.z;
  outvec->x = temp.x;
  outvec->y = temp.y;
}


vector vec_add (vector vec1, vector vec2) //adds vec1 and vec2
{
  return vector((vec1.x + vec2.x), (vec1.y + vec2.y), (vec1.z + vec2.z));
}

vector vec_sub (vector vec1, vector vec2) //adds vec1 and vec2
{
  return vector((vec1.x - vec2.x), (vec1.y - vec2.y), (vec1.z - vec2.z));
}

vector vec_div (vector vec, float div)
{
  return vector((vec.x / div), (vec.y / div), (vec.z / div));
}

vector vec_mul (vector vec, float mul)
{
  return vector((vec.x * mul), (vec.y * mul), (vec.z * mul));
}

vector vec_project(vector vec)  //converts pitch,yaw,dir to x,y,z
{
  return vector(
    (vec.z * (-(float)sin(vec.y) * (float)cos(vec.x))),
    (vec.z * ((float)sin(vec.x))),
    (vec.z * ((float)cos(vec.x) * (float)cos(vec.y))));
}

vector vec_set(float val)
{
  return vector(val, val, val);
}

char vec_is(vector vec)
{
  return ((vec.x) || (vec.y) || (vec.z));
}

vector vec_neg(vector vec)
{
  return vector(-vec.x,-vec.y,-vec.z);
}

vector vec_cross(vector vec1, vector vec2)
{
  return vector(
    (vec1.y * vec2.z - vec2.y * vec1.z),
    (vec1.z * vec2.x - vec2.z * vec1.x),
    (vec1.x * vec2.y - vec2.x * vec1.y));
}

vector vec_normal (vector vec)
{
  float len = (float)sqrt(vec.x * vec.x + vec.y * vec.y + vec.z * vec.z);
  if (len)
    return vec_div(vec,len);
  else
    return vec_set(0);
}

float vec_dot(vector vec1, vector vec2) //you MUST normalize both
{                                               //vectors before you use this
  return (vec1.x * vec2.x + vec1.y * vec2.y + vec1.z * vec2.z);
}

