/*

    File: vector.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#if !defined(VECTOR_H)
#define VECTOR_H

#include <math.h>
#include <stdlib.h>	//for random
#define M_PI            3.14159265358979323846f

class vector {
public:
	vector(float vx, float vy, float vz) {x=vx;y=vy;z=vz;}
	vector() {x=y=z=0;}

	union {
		struct {
			float x, y, z;
		};
		float v[3];
	};


	float length(void) {return (float)sqrt(x*x+y*y+z*z);}
	void normalize(void) {float l; if (l=length()) scale(1.f/l); return;}
	void scale(float f) {x*=f;y*=f;z*=f;}
	void add(vector *a, vector *b) {x=a->x+b->x;y=a->y+b->y;z=a->z+b->z;}
	void add(vector *a) {x+=a->x;y+=a->y;z+=a->z;}
	void sub(vector *a, vector *b) {x=a->x-b->x;y=a->y-b->y;z=a->z-b->z;}
	void sub(vector *a) {x-=a->x;y-=a->y;z-=a->z;}
	void cross(vector *a, vector *b) {x=a->y*b->z-a->z*b->y;y=a->z*b->x-a->x*b->z;z=a->x*b->y-a->y*b->x;}
};

#define DegtoRad(x) (x*M_PI/180)
#define RadtoDeg(x) (x*180/M_PI)
#define Random() ((float)rand() / (float)RAND_MAX)
#define RandomNum() (Random() * 2 - 1)

vector vec_rot_x (vector vec, float rot);
vector vec_rot_y (vector vec, float rot);
vector vec_rot_z (vector vec, float rot);
void rotate (vector vec, vector *outvec, vector rot);
vector vec_add (vector vec1, vector vec2); //adds vec1 and vec2
vector vec_sub (vector vec1, vector vec2); //adds vec1 and vec2
vector vec_div (vector vec, float div);
vector vec_mul (vector vec, float mul);
vector vec_project(vector vec);  //converts pitch,yaw,dir to x,y,z
vector vec_set(float val);
char vec_is(vector vec);
vector vec_neg(vector vec);
vector vec_cross(vector vec1, vector vec2);
vector vec_normal (vector vec);
float vec_dot(vector vec1, vector vec2); //you MUST normalize both

#endif